;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; PDF Parser
;;; 
;;; (C) Arthur Lemmens, 2004
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defpackage :parse-pdf)

(in-package :parse-pdf)

;; DO: Warn about encrypted files.

(defconstant +external-format+ #-(or lispworks clisp allegro) :default
   #+(and allegro mswindows) (excl:crlf-base-ef :1252) ;;:1252-base
   #+(and allegro unix) :default
   #+lispworks '(:latin-1 :eol-style :lf)
   #+clisp :unix)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Dictionaries and streams
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass dictionary ()
  ((properties :initarg :properties :initform '() :accessor properties)))

(defun raw-dictionary-value (key dictionary)
  (getf (properties dictionary) key))

(defun (setf raw-dictionary-value) (value key dictionary)
  (setf (getf (properties dictionary) key)
        value))

(defun map-dictionary (function dictionary)
  "Executes a binary function for all key/value pairs in dictionary."
  (loop for (key value) on (properties dictionary) by #'cddr
        do (funcall function key value)))

(defun make-dictionary-properties (plist)
  plist)

(defclass pdf-stream (dictionary)
  ((contents :accessor contents :initform '() :initarg :contents
             :documentation "A list of strings.")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Cross-reference table
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass cross-reference-table ()
  ((sections :initarg :sections :initform '() :accessor sections
             :documentation "A list of cross-reference-sections (most recent first).")))

(defclass cross-reference-section ()
  ((subsections :initarg :subsections :initform '() :accessor subsections)
   (trailer :initarg :trailer :accessor trailer)))

(defclass cross-reference-subsection ()
  ((first-entry :initarg :first-entry :accessor first-entry)
   (nr-entries  :initarg :nr-entries  :accessor nr-entries)
   (entries     :initarg :entries
                :initform (make-array 100 :adjustable t :fill-pointer 0)
                :accessor entries)))

(defclass cross-reference-entry ()
  ((generation-nr :initarg :generation-nr :accessor generation-nr)))

(defclass free-cross-reference-entry (cross-reference-entry)
  ((next-free-object-nr :initarg :next-free-object-nr :accessor next-free-object-nr)))

(defclass in-use-cross-reference-entry (cross-reference-entry)
  ((byte-offset :initarg :byte-offset :accessor byte-offset)))


(defclass trailer (dictionary)
  ())

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Document
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *document*)

(defclass catalog (indirect-object)
  ())

(defclass document ()
  ((objects :accessor objects :initform (make-hash-table)
            :documentation "A hash-table mapping an object number to a list of indirect-objects.")
   (file :initarg :file :reader file)
   (stream :initarg :stream :reader document-stream)
   (cross-reference-table :initarg :cross-reference-table :accessor cross-reference-table)
   (catalog :initarg :catalog :initform nil :accessor catalog)
   (info :initarg :info :initform nil :accessor info)
   (encrypt :initarg :encrypt :initform nil :accessor encrypt)))

#|
   ;; LATER
   (root-page :accessor root-page :initform nil)
   (catalog :accessor catalog :initform nil)
   (outline-root :accessor outline-root :initform nil)
   (named-refs :accessor named-refs :initform (make-hash-table :test #'equal))
   (fonts :accessor fonts :initform '())
   (gstates :accessor gstates :initform '())
   (encodings :accessor encodings :initform '())
   (last-object-number :accessor last-object-number :initform 0)
   (docinfo :accessor docinfo :initform nil)
   (author :accessor author :initarg :author :initform nil)
   (title  :accessor title :initarg :title :initform nil)
   (keywords :accessor keywords :initarg :keywords :initform nil)
   (subject :accessor subject :initarg :subject :initform nil)))
|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Indirect objects
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass indirect-object ()
  ((object-nr :accessor object-nr :initarg :object-nr)
   (generation-nr :accessor generation-nr :initform 0 :initarg :generation-nr)
   (contents :accessor contents :initform nil :initarg :contents)
   (document :initarg :document :initform *document*)
   (renumbered-p :initform nil :accessor renumbered-p)))

(defclass indirect-reference ()
  (;; The original object- and generation-number
   (object-nr :accessor object-nr :initarg :object-nr)
   (generation-nr :accessor generation-nr :initform 0 :initarg :generation-nr)
   ;;
   (object :accessor object :documentation "The referenced object")))

(defmethod print-object ((object indirect-object) stream)
  (print-unreadable-object (object stream :type t :identity nil)
    (with-slots (object-nr generation-nr contents)
        object
      (format stream "~D.~D ~A" object-nr generation-nr contents))))

(defmethod print-object ((object indirect-reference) stream)
  (print-unreadable-object (object stream :type t :identity nil)
    (with-slots (object-nr generation-nr)
        object
      (format stream "~D.~D" object-nr generation-nr))))

(defun dictionary-value (key dictionary document)
  (let ((value (raw-dictionary-value key dictionary)))
    (if (typep value 'indirect-reference)
        (contents (resolve-indirect-reference value document))
      value)))

(defun array-element (index array document)
  (let ((value (aref array index)))
    (if (typep value 'indirect-reference)
        (contents (resolve-indirect-reference value document))
      value)))


(defun resolve-indirect-reference (reference document &key (errorp nil))
  (if (slot-boundp reference 'object)
      (slot-value reference 'object)
    (with-slots (object-nr generation-nr)
        reference
      (setf (slot-value reference 'object)
            (resolve-indirect-reference* object-nr generation-nr document :errorp errorp)))))

(defun resolve-indirect-reference* (object-nr generation-nr document &key (errorp nil))
  ;; Find it in the hash-table or read it from the file.
  (or (find-indirect-object object-nr generation-nr document)
      (let* ((entry (find-used-cross-reference-entry object-nr generation-nr document))
             (position (and entry (byte-offset entry))))
        (or (and position
                 (read-indirect-object (document-stream document) position document))
            (and errorp
                 (error "Can't find indirect object with object-number ~D and generation-number ~D"
                        object-nr generation-nr))))))

(defun find-indirect-object (object-nr generation-nr document)
  (some (lambda (object) 
          (and (= (generation-nr object) generation-nr)
               object))
        (gethash object-nr (objects document))))


(defmethod find-used-cross-reference-entry (object-nr generation-nr document)
  (some (lambda (section)
          (some (lambda (subsection)
                  (and (object-nr-in-subsection-p object-nr subsection)
                       (let ((entry (aref (entries subsection)
                                          (- object-nr (first-entry subsection)))))
                         (and (typep entry 'in-use-cross-reference-entry) entry))))
                (subsections section)))
        (sections (cross-reference-table document))))

(defun object-nr-in-subsection-p (object-nr subsection)
  (<= 0
      (- object-nr (first-entry subsection))
      (1- (nr-entries subsection))))

#+later
(defun merge-documents (documents)
  ;; DO: FINISH THIS
  'do-this)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Errors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-condition pdf-parse-error ()
  ((stream :initarg :stream :reader pdf-parse-error-stream)
   (message :initarg :message :reader pdf-parse-error-message))
  (:report (lambda (condition stream)
             (format stream "~&Error while reading PDF document.~%~A~%~A~%"
                     (pdf-parse-error-message condition)
                     (pdf-parse-error-stream condition)))))

(defun unexpected-character-error (char input-stream)
  (cerror "Ignore the character and continue."
          'pdf-parse-error
          :stream input-stream
          :message (format nil "Unexpected character ~S." char)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Low level stuff
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun white-char-p (c)
  (find c '(#\Space #\Newline #\Return #\Tab #\Null #\Page)))

(defun octal-digit-char-p (char)
  (find char "01234567"))

(defun hex-digit-char-p (char)
  (find char "0123456789ABCDEF"))

(defun numeric-char-p (char)
  (find char "0123456789-."))

(defun name-char-p (char)
  (and (<= #x21 (char-code char) #x7E)
       (not (find char "%()<>[]{}/#"))))

;;
;; Skipping characters
;;

(defun eat-char (expected-char input-stream)
  (let ((char (read-char input-stream)))
    (unless (char= char expected-char)
      (cerror "Ignore the character and continue."
              'pdf-parse-error
              :stream input-stream
              :message (format nil "Unexpected character ~S (expected ~S)."
                               char expected-char)))))

(defun eat-chars (chars input-stream)
  (loop for char across chars
        do (eat-char char input-stream)))

(defun eat-keyword (keyword input-stream)
  (skip-whitespace input-stream t)
  (eat-chars keyword input-stream)
  (skip-whitespace input-stream t))

(defun skip-whitespace (input-stream eof-error-p &key (skip-comments t))
  (loop
   (let ((char (read-char input-stream eof-error-p)))
     (cond ((null char)
            (return :eof))
           ((and skip-comments (eql char #\%))
            (loop for c = (read-char input-stream nil)
                  until (or (null c) (eq c #\Newline))))
           ((not (white-char-p char))
            (unread-char char input-stream)
            (return char))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; PDF Basic objects (see Chapter 4)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun read-object (input-stream document &optional (eof-error-p t))

  "Returns one of the following PDF objects: boolean (:true or :false),
number (Lisp number), string (Lisp string), name (Lisp symbol in the PDF
package), array (Lisp vector), dictionary (Lisp property list), stream
(Lisp pdf-stream) or null (Lisp NIL). When EOF-ERRORP is nil, it returns
:eof for the end of the stream (otherwise it signals an error)." 

  (skip-whitespace input-stream eof-error-p)
  (let ((char (peek-char nil input-stream eof-error-p)))
    (cond ((numeric-char-p char)
           (read-number input-stream))
          ((eql char #\()
           (eat-char #\( input-stream)
           (read-pdf-string input-stream))
          ((eql char #\/)
           (eat-char #\/ input-stream)
           (read-name input-stream))
          ((eql char #\[)
           (eat-char #\[ input-stream)
           (read-array input-stream document))
          ((eql char #\<) 
           (eat-char #\< input-stream)
           (let ((next-char (peek-char nil input-stream)))
             (if (char= next-char #\< ) 
                 (progn (eat-char #\< input-stream)
                   (read-dictionary-or-stream input-stream document))
               (read-hex-string input-stream))))
          ((eql char #\t)
           (eat-chars "true" input-stream)
           :true)
          ((eql char #\f)
           (eat-chars "false" input-stream)
           :false)
          ((eql char #\n)
           (eat-chars "null" input-stream)
           nil)
          (t (unexpected-character-error char input-stream)))))

(defun read-number (input-stream)
  (read-from-string
   (with-output-to-string (s)
     (loop for char = (read-char input-stream nil nil)
           until (or (null char) (not (numeric-char-p char)))
           do (write-char char s)
           finally (when char
                     (unread-char char input-stream))))))

;;
;; Strings
;;

(defun read-pdf-string (input-stream)
  ;; DO: Deal with encodings.
  ;; #\( has already been read from the stream.
  (with-output-to-string (s)
    ;; The #\( and #\) delimiters are not required to match. 
    ;; Just stop at the first non-escaped #\) character.
      (loop for char = (read-char input-stream)
            until (char= char #\))
            do (if (char= char #\\)
                   ;; Handle escaped characters
                   (let ((next-char (read-char input-stream)))
                     (case next-char 
                       ( #\newline
                         ;; ignore
                         )
                       ( #\n
                         (write-char #\newline s))
                       ( #\r
                         (write-char #\return s))
                       ( #\t
                         (write-char #\tab s))
                       ( #\b
                         (write-char #\backspace s))
                       ( #\f
                         (write-char #\formfeed s))
                       ((#\( #\) #\\)
                        (write-char next-char s))
                       (t (cond ((octal-digit-char-p next-char)
                                 (write-char (read-octal-character-code next-char input-stream) s))
                                (t (write-char char s)
                                   (write-char next-char s))))))
                 (write-char char s)))))

(defun read-octal-character-code (first-digit input-stream)
  (let* ((digits (list first-digit))
         (char (peek-char nil input-stream)))
    (when (octal-digit-char-p char)
      (push char digits)
      (read-char input-stream)
      (let ((char (peek-char nil input-stream)))
        (when (octal-digit-char-p char)
          (read-char input-stream)
          (push char digits))))
    (flet ((digit-value (digit)
             (if digit
                 (- (char-code digit) #.(char-code #\0))
               0)))
      (code-char (+ (digit-value (first digits))
                    (* 8 (digit-value (second digits)))
                    (* 64 (digit-value (third digits))))))))

(defun read-hex-string (input-stream)
  (with-output-to-string (s)
    (loop
     (let ((digit-1 (char-upcase (read-char input-stream))))
       (cond ((hex-digit-char-p digit-1)
              (let ((digit-2 (char-upcase (read-char input-stream))))
                (cond ((hex-digit-char-p digit-2)
                       (write-char (code-char (parse-hex digit-1 digit-2))
                                   s))
                      ((char= digit-2 #\>)
                       (write-char (code-char (parse-hex digit-1 #\0))
                                   s)
                       (return))
                      (t (error 'pdf-parse-error
                                :stream input-stream
                                :message (format nil "Unexpected character ~S in hexadecimal string."
                                                 digit-2))))))
             ((char= digit-1 #\>)
              (return))
             (t (error 'pdf-parse-error
                       :stream input-stream
                       :message (format nil "Unexpected character ~S in hexadecimal string."
                                        digit-1))))))))

(defun parse-hex (digit-1 digit-2)
  (flet ((parse-digit (digit)
           (- (char-code digit)
              (char-code (if (char<= #\0 digit #\9) #\0 #\A)))))
    (+ (* 16 (parse-digit (char-upcase digit-1)))
       (parse-digit (char-upcase digit-2)))))

;;
;; Names
;;

(defun read-name (input-stream)
  (intern (with-output-to-string (s)
            ;; (write-char #\/ s) ; Don't include the /.
            (loop (let ((char (read-char input-stream nil nil)))
                    (cond ((null char) (return))
                          ((eql char #\#)
                           (let ((digit-1 (read-char input-stream))
                                 (digit-2 (read-char input-stream)))
                             (unless (and (hex-digit-char-p digit-1)
                                          (hex-digit-char-p digit-2))
                               (error 'pdf-parse-error
                                      :stream input-stream
                                      :message "Illegal hexadecimal escape sequence in PDF name."))
                             (write-char (code-char (parse-hex digit-1 digit-2))
                                         s)))
                          ((name-char-p char)
                           (write-char char s))
                          (t (unread-char char input-stream)
                             (return))))))
          #.*package*))


;;
;; Arrays
;;

(defun read-array (input-stream document)
  ;; #\[ has already been read
  (let ((stack '()))
    (loop (skip-whitespace input-stream t)
          (let ((char (peek-char nil input-stream)))
            (case char
              ( #\]
                (eat-char #\] input-stream)
                (return))
              ( #\R
                (eat-char #\R input-stream)
                ;; Reduce last two numbers to indirect-reference
                (let ((generation-number (pop stack))
                      (object-number (pop stack)))
                  (assert (integerp generation-number))
                  (assert (integerp object-number))
                  (push (make-instance 'indirect-reference
                                       :object-nr object-number
                                       :generation-nr generation-number)
                        stack)))
              (otherwise
               (push (read-object input-stream document) stack)))))
    (make-array (length stack)
                :initial-contents (nreverse stack))))


;;
;; Dictionaries and streams
;;

(defun read-dictionary-properties (input-stream document)
  (let ((plist '()))
    (loop (skip-whitespace input-stream t)
          (let ((char (peek-char nil input-stream)))
            (case char
              ( #\>
                (eat-chars ">>" input-stream)
                (return))
              ( #\R
                (eat-char #\R input-stream)
                ;; Reduce last two numbers to indirect-object
                (let ((generation-number (pop plist))
                      (object-number (pop plist)))
                  (assert (integerp generation-number))
                  (assert (integerp object-number))
                  (push (make-instance 'indirect-reference
                                       :object-nr object-number
                                       :generation-nr generation-number)
                        plist)))
              (otherwise
               (push (read-object input-stream document t)
                     plist)))))
    (nreverse plist)))

(defun read-dictionary (input-stream document)
  (let ((properties (read-dictionary-properties input-stream document)))
    (make-instance 'dictionary 
                   :properties (make-dictionary-properties properties))))

(defun read-dictionary-or-stream (input-stream document)
  (let ((properties (read-dictionary-properties input-stream document)))
    ;; Check if dictionary is followed by a stream
    (skip-whitespace input-stream nil)
    (let ((char (peek-char nil input-stream nil nil)))
      (cond ((eql char #\s)
             ;; Don't use EAT-KEYWORD here, because it may eat too many newlines!
             (eat-chars "stream" input-stream) 
             (read-line input-stream)
             (read-pdf-stream properties input-stream document))
            (t (make-instance 'dictionary 
                              :properties (make-dictionary-properties properties)))))))

(defun read-pdf-stream (plist input-stream document)
  (let ((length (getf plist '|Length|)))
    (when (typep length 'indirect-reference)
      (let ((saved-filepos (file-position input-stream)))
        (setq length (contents (resolve-indirect-reference length document)))
        (file-position input-stream saved-filepos)))
    (assert (integerp length))
    (let ((contents (loop repeat (ceiling length array-total-size-limit)
                          for buffer-size = (min length array-total-size-limit)
                          ;; while (plusp buffer-size)
                          collect (let* ((buffer (make-string buffer-size))
                                         (bytes-read (read-sequence buffer input-stream)))
                                    (when (< bytes-read buffer-size)
                                      (error 'pdf-parse-error
                                             :stream input-stream
                                             :message "Unexpected end of PDF-stream."))
                                    buffer)
                          do (decf length buffer-size))))
      (eat-keyword "endstream" input-stream)
      (make-instance 'pdf-stream
                     :properties (make-dictionary-properties plist)
                     :contents contents))))

;;
;; Integers
;;

(defun read-integer (input-stream)
  (let ((object (read-object input-stream nil)))
    (unless (integerp object)
      (error 'pdf-parse-error 
             :stream input-stream
             :message "Integer expected."))
    object))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Indirect objects
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun read-indirect-object (input-stream file-position document)
  (file-position input-stream file-position)
  (let* ((object-number (read-integer input-stream))
         (generation-number (read-integer input-stream)))
    (eat-keyword "obj" input-stream)
    (let ((object (read-object input-stream document)))
      ;; Some producers forget the "endobj" at the end of a stream
      ;; Let's try to be tolerant.
      (skip-whitespace input-stream t)
      (when (char= #\e (peek-char nil input-stream))
        (eat-keyword "endobj" input-stream))
      ;;
      (let ((indirect-object (make-instance 'indirect-object
                                            :object-nr object-number
                                            :generation-nr generation-number
                                            :contents object
                                            :document document)))
        ;; Add indirect-object to document's hash-table.
        (push indirect-object
              (gethash object-number (objects document)))
        ;;
        indirect-object))))

;;
;; Loading all indirect objects that are still in use.
;;

#| OLD VERSION
(defun load-all-indirect-objects (input-stream document)
  ;; This may fail for cross-reference tables that reference objects that
  ;; don't exist anymore (they shouldn't do that of course, but Acrobat Reader
  ;; has no problems parsing such files, so we should accept them too).
  (let ((table (cross-reference-table document)))
    (dolist (section (sections table))
      (dolist (subsection (subsections section))
        (loop for entry across (entries subsection)
              for object-nr from (first-entry subsection)
              for generation-nr = (generation-nr entry)
              when (and (typep entry 'in-use-cross-reference-entry)
                        (not (find-indirect-object object-nr generation-nr document)))
              do (read-indirect-object input-stream (byte-offset entry) document))))))
|#

(defun load-all-indirect-objects (document)
  (load-indirect-objects (catalog document) document)
  (load-indirect-objects (info document) document)
  (load-indirect-objects (encrypt document) document))


(defmethod load-indirect-objects ((object t) document)
  ;; Default: ignore
  object)

(defmethod load-indirect-objects ((object indirect-reference) document)
  (unless (slot-boundp object 'object)
    (load-indirect-objects (resolve-indirect-reference object document) document)))

(defmethod load-indirect-objects ((object indirect-object) document)
  (load-indirect-objects (contents object) document))

(defmethod load-indirect-objects ((object dictionary) document)
  (map-dictionary (lambda (key value)
                    (declare (ignore key))
                    (load-indirect-objects value document))
                  object))

(defmethod load-indirect-objects ((object vector) document)
  (loop for x across object
        do (load-indirect-objects x document)))

;;
;; MAP-INDIRECT-OBJECTS
;;

(defun map-indirect-objects (function document)
  (loop for objects being the hash-value of (objects document)
        do (loop for object in objects
                 do (funcall function object))))

;;
;; Renumber objects
;;

(defun renumber-all-indirect-objects (document &key (start 1))
  "Gives all indirect objects a consecutive numbering.
Returns the first unused object-number."
  (let ((object-nr start))
    (map-indirect-objects (lambda (object) 
                            (setf (object-nr object) object-nr
                                  (generation-nr object) 0)
                            (incf object-nr))
                          document)
    object-nr))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Higher level stuff
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun parse-pdf (file)
  (with-open-file (s file
                     :direction :input
                     :external-format +external-format+)
    (parse-pdf-stream s)))

(defun parse-pdf-stream (input-stream)
  (let* ((cross-reference (read-cross-reference-table input-stream
                                                      (find-cross-reference-start input-stream)))
         (document (make-instance 'document
                                  :file (truename input-stream)
                                  :stream input-stream
                                  :cross-reference-table cross-reference))
         (trailer (trailer (first (sections cross-reference)))))
    (setf (catalog document) (raw-dictionary-value '|Root| trailer)
          (info document) (raw-dictionary-value '|Info| trailer)
          (encrypt document) (raw-dictionary-value '|Encrypt| trailer))
    (load-all-indirect-objects document)
    (renumber-all-indirect-objects document)
    document))

(defun read-cross-reference-table (input-stream position)
  (let ((sections '()))
    (loop 
     (let* ((subsections (read-cross-reference-subsections input-stream position))
            (trailer (read-trailer input-stream)))
       (push (make-instance 'cross-reference-section
                            :subsections subsections
                            :trailer trailer)
             sections)
       (let ((prev-position (raw-dictionary-value '|Prev| trailer)))
         (if prev-position
             ;; Keep reading cross-reference sections
             (setq position prev-position)
           ;; Done.
           (return (make-instance 'cross-reference-table
                                  :sections (nreverse sections)))))))))
          
 
(defun read-cross-reference-subsections (input-stream position)
  (file-position input-stream position)
  (eat-keyword "xref" input-stream)
  (let ((subsections '()))
    (loop
     (let ((char (peek-char nil input-stream)))
       (cond ((char= char #\t)
              ;; Start of trailer
              (return (nreverse subsections)))
             (t (push (read-cross-reference-subsection input-stream)
                      subsections)))))))

(defun read-cross-reference-subsection (input-stream)      
  (let ((first-entry (read-integer input-stream))
        (nr-entries (read-integer input-stream))
        (entries (make-array 5 :adjustable t :fill-pointer 0)))
    (loop repeat nr-entries
          do (let ((entry (read-cross-reference-entry input-stream)))
               (vector-push-extend entry entries)))
    (make-instance 'cross-reference-subsection
                   :first-entry first-entry
                   :nr-entries nr-entries
                   :entries entries)))

(defun read-cross-reference-entry (input-stream)
  (let ((first (read-integer input-stream)))
    (skip-whitespace input-stream nil)
    (let ((second (read-integer input-stream)))
      (skip-whitespace input-stream nil)
      (let ((third (read-char input-stream)))
        (skip-whitespace input-stream nil)
        (if (char= third #\n)
            (make-instance 'in-use-cross-reference-entry
                           :byte-offset first
                           :generation-nr second)
          (make-instance 'free-cross-reference-entry
                         :next-free-object-nr first
                         :generation-nr second))))))

(defun read-trailer (input-stream)
  (eat-keyword "trailer" input-stream)
  (skip-whitespace input-stream t)
  (eat-chars "<<" input-stream)
  (change-class (read-dictionary input-stream nil) ; document irrelevant here
                'trailer))


;;
;;
;;

(defun find-cross-reference-start (input-stream)
  (let ((file-length (file-length input-stream))
        (buffer (make-string 100)))
    (file-position input-stream (- file-length 100))
    (read-sequence buffer input-stream)
    (let ((position (search "startxref" buffer)))
      (unless position
        (error 'pdf-parse-error
               :stream input-stream
               :message "Can't find start address of cross-reference table."))
      (parse-integer buffer :start (+ position 10) :junk-allowed t))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Write-document
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter +pdf-header+ "%PDF-1.4")

(defvar *compress-streams* nil)
(defvar *min-size-for-compression* 100000)
(defvar *xrefs* nil)
(defvar *pdf-stream* nil)

 

(defmethod write-object ((obj dictionary) stream &optional root-level)
  (declare (ignorable root-level))
  (write-string "<< " stream)
  (map-dictionary (lambda (key val)
                    (write-object key stream)
                    (write-char #\Space stream)
                    (write-object val stream)
                    (write-char #\Newline stream))
                  obj)
  (write-line " >>" stream))

(defmethod write-object ((obj pdf-stream) stream &optional root-level)
  (declare (ignorable root-level))
  #+later (when (and *compress-streams* (not (no-compression obj))
                     (> (length (contents obj)) *min-size-for-compression*))
            (setf (contents obj) (compress-string (content obj)))
            (let ((filter (get-dict-value obj "/Filter")))
              (if filter
                  (change-dict-value obj "/Filter" (vector "/FlateDecode" filter))
                (add-dict-value obj "/Filter" "/FlateDecode"))))
  (call-next-method) ; write dictionary
  (write-line "stream" stream)
  (loop for buffer in (contents obj)
        do
        #+pdf-binary(write-sequence buffer stream)
        #-pdf-binary(if (stringp buffer)
                        (write-sequence buffer stream)
                      (loop for c across buffer do
                            (write-char (code-char c) stream))))
  (write-char #\newline stream)
  (write-string "endstream" stream))

(defmethod write-object ((obj indirect-reference) stream &optional root-level)
  (declare (ignorable root-level))
  (if (object obj)
      (write-object-reference (object obj) stream)
    (write-object nil stream)))

(defmethod write-object ((obj indirect-object) stream &optional root-level)
  (if root-level
    (progn
      (vector-push-extend (format nil "~10,'0d ~5,'0d n "
				  (file-position stream)
                                  (generation-nr obj))
			  *xrefs*)
      (format stream "~d ~d obj~%" (object-nr obj) (generation-nr obj))
      (write-object (contents obj) stream)
      (write-char #\newline stream)
      (write-line "endobj" stream))
    (write-object-reference obj stream)))

(defun write-object-reference (obj stream)
  (format stream "~d ~d R " (object-nr obj) (generation-nr obj)))

(defmethod write-object ((list list) stream &optional root-level)
  (declare (ignorable root-level))
  (dolist (obj list)
    (write-object obj stream)
    (write-char #\Space stream)))

(defmethod write-object ((obj string) stream &optional root-level)
  (declare (ignorable root-level))
  ;; Add PDF's delimiters to the string.
  (write-char #\( stream)
  (loop for char across obj
        do (case char
             (#\( (write-string "\\(" stream))
             (#\) (write-string "\\)" stream))
             (#\\ (write-string "\\\\" stream))
             (#\newline (write-string "\\n" stream))
             (#\return  (write-string "\\r" stream))
             (#\tab (write-string "\\t" stream))
             (#\backspace (write-string "\\b" stream))
             (#\formfeed (write-string "\\f" stream))
             (otherwise 
              (cond ((name-char-p char)
                     (write-char char stream))
                    (t 
                     ;; Write octal escape sequence
                     (write-char #\\ stream)
                     (let ((code (char-code char)))
                       (write (floor code 64) :stream stream)
                       (write (floor (mod code 64) 8) :stream stream)
                       (write (mod code 8) :stream stream)))))))
  (write-char #\) stream))


(defmethod write-object ((obj null) stream &optional root-level)
  (write-string "null" stream))

(defmethod write-object ((obj symbol) stream &optional root-level)
  (declare (ignorable root-level))
  (cond ((keywordp obj)
         ;; Must be :true or :false
         (format stream "~(~A~)" obj))
        (;; Write out a name
         (write-char #\/ stream)
         (loop for char across (symbol-name obj)
               do (cond ((name-char-p char)
                         (write-char char stream))
                        (t
                         ;; Use hex escape sequence
                         (write-char #\# stream)
                         (write (floor (char-code char) 256) :stream stream)
                         (write (mod (char-code char) 256) :stream stream)))))))

(defmethod write-object ((obj function) stream &optional root-level)
  (declare (ignorable root-level))
  (write-object (funcall obj) stream))

(defmethod write-object ((obj t) stream &optional root-level)
  ;; For numbers
  (declare (ignorable root-level))
  (format stream "~a" obj))

(defmethod write-object ((array array) stream &optional root-level)
  (declare (ignorable root-level))
  (write-string "[ " stream)
  (loop for obj across array do
	(write-object obj stream)
	(write-char #\Space stream))
  (write-char #\] stream))


(defmethod write-document ((s stream) &optional (document *document*))
  (let ((*xrefs* (make-array 10 :adjustable t :fill-pointer 0))
        (*pdf-stream* s))
    (with-standard-io-syntax
      #+later(process-outlines document)
      (vector-push-extend "0000000000 65535 f " *xrefs*)
      (write-line +pdf-header+ s)
      (write-document-objects s document)
      (let ((startxref (file-position s)))
        (format s "xref~%0 ~d~%" (length *xrefs*))
        (loop for xref across *xrefs*
              do (write-line xref s))
        (format s "trailer~%<< /Size ~d~% " (length *xrefs*));(1- (length (objects document))))
        (when (catalog document)
          (format s "/Root ")
          (write-object (catalog document) s))
        (when (info document)
          (format s "/Info ")
          (write-object (info document) s))
        (when (encrypt document)
          (format s "/Encrypt ")
          (write-object (encrypt document) s))
        (format s "~%>>~%startxref~%~d~%%%EOF~%" startxref)))))

(defun write-document-objects (stream document)
  (let ((objects (loop for objects being the hash-value of (objects document)
                       append objects)))
    (setq objects (sort objects (lambda (a b)
                                  (or (< (object-nr a) (object-nr b))
                                      (and (= (object-nr a) (object-nr b))
                                           (< (generation-nr a) (generation-nr b)))))))
    (loop for obj in objects
          for first = t then nil
          if obj do (write-object obj stream t)
          else do (unless first (vector-push-extend "0000000000 00001 f " *xrefs*)))))

(defmethod write-document ((filename pathname) &optional (document *document*))
   (with-open-file (s filename :direction :output :if-exists :supersede
		      :external-format +external-format+)
     (write-document s document)))

(defmethod write-document ((filename string) &optional (document *document*))
  (write-document (pathname filename) document))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Testing
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun copy-document (input-file output-file)
  (let ((document (parse-pdf input-file)))
    (write-document output-file document)
    document))

(defun copy-examples ()
  (loop for i from 1 to 8
        do (copy-document (format nil "c:/tmp/ex~d.pdf" i)
                          (format nil "c:/tmp/Copy of ex~D.pdf" i))))
  
(defun copy-all-documents-in-directory (directory)
  (loop for file in (directory directory)
        for i from 1
        do (let ((name (pathname-name file))
                 (type (pathname-type file)))
             (when (and (string-equal type "PDF")
                        (or (< (length name) 7)
                            (mismatch "Copy of" (subseq name 0 7))))
               (let ((output-file (make-pathname :name (format nil "Copy of ~A" name)
                                                 :defaults file)))
                 (unless (probe-file output-file)
                   (format *trace-output* "~&~3D: Copying ~A to ~A~%" i name output-file)
                   (copy-document file output-file)))))))
         

